
%{
#include  <stdio.h>
#define YYSTYPE union stacktype

/*
struct for the nodes of the list 
*/
struct node
{
 int value;
 struct node* next;
};

/*
struct for symtab node
*/
struct symnode
{
 char vname;
 int value;
 struct symnode* next;
};
 
/*
stacktype . The stack contents may be 1)int for NUM and <item>
2)char* for VAR 3) struct mode  for <list>. count gives the number
of elements in the  list
*/
union stacktype
{
 int value;
 struct
 {
	struct node* node_ptr;
	int count;
 }mode;
};

struct node* add_item_to_list (struct node* node_ptr, int value, int count);
int add_list_items (struct node* node_ptr, int count);
int sub_list_items (struct node* node_ptr, int count);
int mul_list_items (struct node* node_ptr, int count);
int div_list_items (struct node* node_ptr, int count);

int car (struct node* node_ptr);
struct node* cdr (struct node* node_ptr);
void print_list (struct node* node_ptr, int count);
void printsymtab ();
int entervar (char vname, int value);
int getvar (char vname);

struct symnode* head;
%}


%token NUMBER
%token VAR
%token CONS
%token LIST
%token CAR
%token CDR
%token SETQ
%token SYMTAB

%%

statement:
			| 	'\n'             
			|	statement list '\n'	{ /*printf ("%d\n", car ($2.mode.node_ptr));*/
											  print_list ($2.mode.node_ptr, $2.mode.count);	
											  printf("myLISP> "); 
											}
			| statement SYMTAB '\n' { printsymtab (); }
			;

list: item							{	$$.mode.node_ptr = add_item_to_list ($$.mode.node_ptr, $1.value, $$.mode.count); $$.mode.count++;}
	 |	list item					{	$$.mode.node_ptr = add_item_to_list ($$.mode.node_ptr, $2.value, $$.mode.count); $$.mode.count++;}
	 | '(' CONS item item ')'	{	$$.mode.node_ptr = add_item_to_list ($$.mode.node_ptr, $3.value, $$.mode.count);
											$$.mode.count++;
											$$.mode.node_ptr = add_item_to_list ($$.mode.node_ptr, $4.value, $$.mode.count);
											$$.mode.count++;
										}
	 | '(' LIST list ')'			{	$$.mode.node_ptr = $3.mode.node_ptr; 
											$$.mode.count = $3.mode.count;
									   }
	 | '(' CDR list ')'			{  $$.mode.node_ptr = cdr ($3.mode.node_ptr); 
											$$.mode.count = $3.mode.count-1;
										}
	 ; 

item:	'(' '+' list ')'		{ $$.value = add_list_items ($3.mode.node_ptr, $3.mode.count); }
	 |	'(' '-' list ')'		{ $$.value = sub_list_items ($3.mode.node_ptr, $3.mode.count); }
	 | '(' '*' list ')'		{ $$.value = mul_list_items ($3.mode.node_ptr, $3.mode.count); }
	 | '(' '/' list ')'		{ $$.value = div_list_items ($3.mode.node_ptr, $3.mode.count); }
	 | '(' CAR list ')'  	{ $$.value = car ($3.mode.node_ptr); }
	 | '(' SETQ VAR item ')'{ $$.value = entervar ($3.value, $4.value);}
	 | NUMBER					{ $$.value = $1.value; }
	 |	VAR						{ $$.value = getvar ($1.value); }

%%

int yyerror (char* s)
{
	printf ("%s\n", s);
	return 1;
}

struct node* add_item_to_list (struct node* head, int value, int count)
{
	int i;
	struct node* new = (struct node*)malloc (sizeof (struct node*));
	struct node* trav = head;
	new->value = value;
	if (count==0)
	{
		return new;
	}
	for (i=0; i<count-1; i++)	trav = trav->next;
	trav->next = new;
	return head;
}

/*
(+ ...)
*/
int add_list_items (struct node* head, int count)
{
	int sum=0,i;
	for (i=0; i<count;i++)
	{
		sum += head->value;
		head = head->next;
	}
	return sum;
}
	 
/*
(- ...)
*/
int sub_list_items (struct node* head, int count)
{
	int dif=head->value, i;
	head=head->next;
	for (i=1; i<count;i++)
	{
		dif -= head->value;
		head = head->next;
	}
	return dif;
}

/*
(* ...)
*/
int mul_list_items (struct node* head, int count)
{
	int pro=1, i;
	for (i=0; i<count;i++)
	{
		pro *= head->value;
		head = head->next;
	}
	return pro;
}

/*
(/ ...)
*/
int div_list_items (struct node* head, int count)
{
	int div=head->value, i;
	head=head->next;
	for (i=1; i<count;i++)
	{
		div /= head->value;
		head = head->next;
	}
	return div;
}

void print_list (struct node* head, int count)
{
 int i;
 printf ("(");
 for (i=0;  i<count; i++)
 {
	printf ("%d ", head->value);
	head=head->next;
 }
 printf (")\n");
}

int car (struct node* head)
{
 return head->value;
}   	

struct node* cdr (struct node* head)
{
 return head->next;
}

/*
enter vars to symbol table
*/
int entervar (char vname, int value)
{
 struct symnode* trav;
 if (head == NULL)
 {
	head = (struct symnode*)malloc (sizeof (struct symnode));
	head->vname = vname;
	head->value = value;
	return value;
 }
 trav = head;
 for ( ;trav->next; trav=trav->next)
 {
	if (trav->vname == vname)
	{
	 printf ("setq: duplicate symbol found");
	 return value;
	}
 }
 trav->next = (struct symnode*)malloc (sizeof (struct symnode*));
 trav=trav->next;
 trav->vname = vname;
 trav->value = value;
 return value;
}

/*
get valueof a variable from symtab 
*/
int getvar (char vname)
{
 struct symnode* trav = head;
 for (;trav;trav=trav->next)
 {
	if(trav->vname==vname)
	 return trav->value;
 }
 printf ("undefined symbol: %c\n", vname);
 return 0;
}

void printsymtab ()
{
 struct symnode* trav = head;
 for (;trav; trav=trav->next)
	printf ("[%c %d] ",trav->vname, trav->value);
 printf ("\nmyLISP> ");
}